var Uploader = Class.create({

    uploadForm: null,
    uploadView: null,
    progressView: null,

    uploadQueue: null,
    commandChannel: null,
    progressMeter: null,
    progressListener: null,

    sessionId: null,
    interrupted: false,
    transferError: null,
    listener: null,

    _constructor: function(config) {
        this.uploadForm = new UploadForm(config.uploadFormUrl);
        this.uploadForm.setForms(config.uploadForms || []);

        this.uploadView =  new UploadView(new FormFileChooser());

        this.progressMeter = new ProgressMeter(config);
        this.progressView = new ProgressView({
            onCancelUpload: this._handleUploadCancel.bind(this)
        });
        this.progressListener = {
            onProgress: this._handleProgressProgress.bind(this),
            onComplete: this._handleProgressComplete.bind(this),
            onCancel: this._handleProgressCancel.bind(this),
            onTimeout: this._handleTimeout.bind(this),
            onError: this._handleError.bind(this)
        };
        this.commandChannel = new ScriptChannel({timeout: config.connectionTimeout, preventCaching: true});
        this.commandChannel.open(config.uploadSessionUrl);
    },

    getTransferError: function() {
        return this.transferError;
    },

    isInterrupted: function() {
        return this.interrupted;
    },

    start: function(listener) {
        this.listener = listener || new Uploader.Listener();
        this.uploadView.show({
            onReady: this._handleReadyState.bind(this),
            onChooseFiles: this._handleChooseFiles.bind(this),
            onRemoveAll: this._handleRemoveAll.bind(this),
            onStartUpload: this._handleStartUpload.bind(this)
        });
    },

    upload: function() {
        this.interrupted = false;
        this.uploadView.hide();
        this.progressView.setVisible(true);
        this.listener.onStart(this);

        this._initialHandshake();
    },

    cancel: function() {
        this.uploadQueue.abort();
        this.progressMeter.interrupt();
        this.progressView.setEnabled(false);
    },

    _initialHandshake: function(data) {
        data = Object.extend(data || {}, {createSession: true, client: this.uploadQueue.client});
        var callback = this._createSession.bind(this);
        this.commandChannel.send(data, callback);
    },

    _createSession: function(response) {
        try {
            this.sessionId = response.id;
            var uploadUrl = response.url;
            this._uploadOperation(uploadUrl);
            this.progressMeter.start(this.progressListener);
        } catch (e) {
            //todo: log error
        }
    },

    _uploadOperation: function(url) {
        this.uploadQueue.send(url, {
            onComplete: this._handleUploadComplete.bind(this),
            onError: this._handleError.bind(this)
        });
    },

    _cancelOperation: function() {
        try {
            this.commandChannel.send({cancelSession: this.sessionId});
        } catch (e) {}
    },

    _handleUploadComplete: function() {
        this.transferError = null;
        this.progressView.setProgress(100);
        this.listener.onComplete(this);
    },

    _handleUploadCancel: function() {
        this.cancel();
        this.interrupted = true;
        this.progressView.setVisible(false);

        this._cancelOperation();
        this.listener.onComplete(this);
    },

    _handleTimeout: function() {
        this.cancel();
        this.uploadView._showConnectionTimeoutMessage();
        this.listener.onTimeout(this);
    },

    _handleError: function(message) {
        this.cancel();
        this.transferError = message;
        this.uploadView._showConnectionErrorMessage(message);
        this.listener.onComplete(this);
    },

    _handleProgressComplete: function() {
        this.progressMeter.finish();
        this.progressView.setEnabled(false);
    },

    _handleProgressProgress: function(percent, statusMessage, timeMessage) {
        this.progressView.setProgress(percent);
        this.progressView.setStatus(statusMessage);
        this.progressView.setTime(timeMessage);
    },

    _handleProgressCancel: function() {
        this.progressMeter.interrupt();
        this.progressView.setEnabled(false);
    },

    _handleReadyState: function(chooser) {
        this.uploadQueue = chooser.createUploadQueue();
    },

    _handleChooseFiles: function(files) {
        for (var i = 0; i < files.length; ++i) {
            if (this.uploadQueue.add(files[i])) {
                this.uploadView.addFile(files[i]);
            }
        }
    },

    _handleRemoveAll: function() {
        this.uploadQueue.clear();
        this.uploadView.clear();
    },

    _handleStartUpload: function() {
        if (this.uploadQueue.isEmpty()) {
            this.uploadView._showNoFilesSelectedMessage();
            return;
        }

        this._retainQueuedWithRemote();

        if (!this.uploadQueue.isEmpty()) {
            var filenames = [];
            var fileItems = this.uploadQueue.fileItems();
            for (var i=0; i < fileItems.length; ++i) {
                filenames.push(fileItems[i].filename);
            }

            this.uploadForm.start(filenames, {
                onStartUpload: this.upload.bind(this),
                onCancelUpload: this.cancel.bind(this)
            });
        }
    },

    _retainQueuedWithRemote: function() {
        var queuedFileItems = this.uploadQueue.fileItems();
        var remoteFilenames = this._listRemoteFilenames();

        for (var i = 0; i < queuedFileItems.length;) {
            var existsOnServer = remoteFilenames.indexOf(queuedFileItems[i].filename) != -1;

            if (!existsOnServer) {
                queuedFileItems.splice(i, 1);
            } else {
                i++;
            }
        }

        if (queuedFileItems.length != 0 && !this.uploadView._confirmOverwriteMessage(queuedFileItems)) {
            for (var i = 0; i < queuedFileItems.length; ++i) {
                var victim = queuedFileItems[i];
                if (this.uploadQueue.remove(victim)) {
                    this.uploadView.removeFile(victim);
                }
            }
        }
    },

    // todo: refactor
    _listRemoteFilenames: function() { return window.listFiles(); }
});
Uploader.Listener = Class.create({
    onStart: function() {},
    onComplete: function() {},
    onTimeout: function() {}
});
Uploader.Messages = {
    NO_FILES_SELECTED: "No files selected.",
    TRANSFER_ERROR: "Transfer error.",
    CONNECTION_TIMEOUT: "Connection timeout.",
    CONFIRM_OVERWRITE: "The target file(s) already exists. Overwrite?"
};
